@@ -15,9 +15,9 @@
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-#include "geo.h"
-#include "pat.h"
-#include "util.h"
+#include "grn_geo.h"
+#include "grn_pat.h"
+#include "grn_util.h"
 
 #include <string.h>
 #include <stdlib.h>
@@ -91,18 +91,18 @@ compute_min_and_max_key(uint8_t *key_base, int diff_bit,
   diff_bit_mask = 0xff >> (diff_bit % 8);
 
   if (diff_byte == sizeof(grn_geo_point)) {
-    if (key_min) { memcpy(key_min, key_base, diff_byte); }
-    if (key_max) { memcpy(key_max, key_base, diff_byte); }
+    if (key_min) { grn_memcpy(key_min, key_base, diff_byte); }
+    if (key_max) { grn_memcpy(key_max, key_base, diff_byte); }
   } else {
     if (key_min) {
-      memcpy(key_min, key_base, diff_byte + 1);
+      grn_memcpy(key_min, key_base, diff_byte + 1);
       key_min[diff_byte] &= ~diff_bit_mask;
       memset(key_min + diff_byte + 1, 0,
              sizeof(grn_geo_point) - diff_byte - 1);
     }
 
     if (key_max) {
-      memcpy(key_max, key_base, diff_byte + 1);
+      grn_memcpy(key_max, key_base, diff_byte + 1);
       key_max[diff_byte] |= diff_bit_mask;
       memset(key_max + diff_byte + 1, 0xff,
              sizeof(grn_geo_point) - diff_byte - 1);
@@ -274,13 +274,13 @@ grn_geo_table_sort_detect_far_point(grn_ctx *ctx, grn_obj *table, grn_obj *index
   grn_gton(geo_key_curr, base_point, sizeof(grn_geo_point));
   *diff_bit = sizeof(grn_geo_point) * 8;
   diff_bit_current = sizeof(grn_geo_point) * 8;
-  memcpy(&point, base_point, sizeof(grn_geo_point));
+  grn_memcpy(&point, base_point, sizeof(grn_geo_point));
   ep = entries;
   inspect_mesh(ctx, &point, *diff_bit, -1);
   while ((tid = grn_pat_cursor_next(ctx, pc))) {
     grn_ii_cursor *ic = grn_ii_cursor_open(ctx, (grn_ii *)index, tid, 0, 0, 1, 0);
     if (ic) {
-      grn_ii_posting *posting;
+      grn_posting *posting;
       grn_gton(geo_key_prev, &point, sizeof(grn_geo_point));
       grn_pat_get_key(ctx, pat, tid, &point, sizeof(grn_geo_point));
       grn_gton(geo_key_curr, &point, sizeof(grn_geo_point));
@@ -553,7 +553,7 @@ grn_geo_table_sort_collect_points(grn_ctx *ctx, grn_obj *table, grn_obj *index,
         if (ic) {
           double d;
           grn_geo_point pos;
-          grn_ii_posting *posting;
+          grn_posting *posting;
           grn_pat_get_key(ctx, pat, tid, &pos, sizeof(grn_geo_point));
           d = grn_geo_distance_rectangle_raw(ctx, base_point, &pos);
           inspect_tid(ctx, tid, &pos, d);
@@ -704,7 +704,7 @@ grn_geo_table_sort(grn_ctx *ctx, grn_obj *table, int offset, int limit,
         while (i < e && (tid = grn_pat_cursor_next(ctx, pc))) {
           grn_ii_cursor *ic = grn_ii_cursor_open(ctx, (grn_ii *)index, tid, 0, 0, 1, 0);
           if (ic) {
-            grn_ii_posting *posting;
+            grn_posting *posting;
             while (i < e && (posting = grn_ii_cursor_next(ctx, ic))) {
               if (offset <= i) {
                 grn_id *v;
@@ -862,7 +862,7 @@ grn_geo_select_in_circle(grn_ctx *ctx, grn_obj *index,
       name_size = grn_obj_name(ctx, domain_object, name, GRN_TABLE_MAX_KEY_SIZE);
       grn_obj_unlink(ctx, domain_object);
     } else {
-      strcpy(name, "(null)");
+      grn_strcpy(name, GRN_TABLE_MAX_KEY_SIZE, "(null)");
       name_size = strlen(name);
     }
     ERR(GRN_INVALID_ARGUMENT,
@@ -1034,7 +1034,7 @@ in_rectangle_data_fill(grn_ctx *ctx, grn_obj *index,
       name_size = grn_obj_name(ctx, domain_object, name, GRN_TABLE_MAX_KEY_SIZE);
       grn_obj_unlink(ctx, domain_object);
     } else {
-      strcpy(name, "(null)");
+      grn_strcpy(name, GRN_TABLE_MAX_KEY_SIZE, "(null)");
       name_size = strlen(name);
     }
     ERR(GRN_INVALID_ARGUMENT,
@@ -1419,8 +1419,8 @@ grn_geo_cursor_area_init(grn_ctx *ctx,
   }
 
   area->current_entry = 0;
-  memcpy(&(area->top_left), &area_top_left, sizeof(grn_geo_point));
-  memcpy(&(area->bottom_right), &area_bottom_right, sizeof(grn_geo_point));
+  grn_memcpy(&(area->top_left), &area_top_left, sizeof(grn_geo_point));
+  grn_memcpy(&(area->bottom_right), &area_bottom_right, sizeof(grn_geo_point));
   grn_gton(area->top_left_key, &area_top_left, sizeof(grn_geo_point));
   grn_gton(area->bottom_right_key, &area_bottom_right, sizeof(grn_geo_point));
 
@@ -1430,7 +1430,7 @@ grn_geo_cursor_area_init(grn_ctx *ctx,
                                  &area_bottom_right,
                                  &data);
   entry->target_bit = data.rectangle_common_bit;
-  memcpy(entry->key, data.rectangle_common_key, sizeof(grn_geo_point));
+  grn_memcpy(entry->key, data.rectangle_common_key, sizeof(grn_geo_point));
   entry->status_flags =
     GRN_GEO_CURSOR_ENTRY_STATUS_TOP_INCLUDED |
     GRN_GEO_CURSOR_ENTRY_STATUS_BOTTOM_INCLUDED |
@@ -1474,8 +1474,8 @@ grn_geo_cursor_open_in_rectangle(grn_ctx *ctx,
 
   cursor->pat = data.pat;
   cursor->index = index;
-  memcpy(&(cursor->top_left), data.top_left, sizeof(grn_geo_point));
-  memcpy(&(cursor->bottom_right), data.bottom_right, sizeof(grn_geo_point));
+  grn_memcpy(&(cursor->top_left), data.top_left, sizeof(grn_geo_point));
+  grn_memcpy(&(cursor->bottom_right), data.bottom_right, sizeof(grn_geo_point));
   cursor->pat_cursor = NULL;
   cursor->ii_cursor = NULL;
   cursor->offset = offset;
@@ -1494,10 +1494,12 @@ grn_geo_cursor_open_in_rectangle(grn_ctx *ctx,
     }
   }
   {
-    const char *minimum_reduce_bit_env;
+    char minimum_reduce_bit_env[GRN_ENV_BUFFER_SIZE];
     cursor->minimum_reduce_bit = 0;
-    minimum_reduce_bit_env = getenv("GRN_GEO_IN_RECTANGLE_MINIMUM_REDUCE_BIT");
-    if (minimum_reduce_bit_env) {
+    grn_getenv("GRN_GEO_IN_RECTANGLE_MINIMUM_REDUCE_BIT",
+               minimum_reduce_bit_env,
+               GRN_ENV_BUFFER_SIZE);
+    if (minimum_reduce_bit_env[0]) {
       cursor->minimum_reduce_bit = atoi(minimum_reduce_bit_env);
     }
     if (cursor->minimum_reduce_bit < 1) {
@@ -1544,7 +1546,7 @@ grn_geo_cursor_entry_next_push(grn_ctx *ctx,
       grn_geo_cursor_area *area;
       area = &(cursor->areas[cursor->current_area]);
       next_entry = &(area->entries[++area->current_entry]);
-      memcpy(next_entry, entry, sizeof(grn_geo_cursor_entry));
+      grn_memcpy(next_entry, entry, sizeof(grn_geo_cursor_entry));
       pushed = GRN_TRUE;
     }
     grn_table_cursor_close(ctx, pat_cursor);
@@ -1578,9 +1580,9 @@ grn_geo_cursor_entry_next(grn_ctx *ctx,
 
   top_left_key = area->top_left_key;
   bottom_right_key = area->bottom_right_key;
-  memcpy(entry,
-         &(area->entries[area->current_entry--]),
-         sizeof(grn_geo_cursor_entry));
+  grn_memcpy(entry,
+             &(area->entries[area->current_entry--]),
+             sizeof(grn_geo_cursor_entry));
   while (GRN_TRUE) {
     grn_geo_cursor_entry next_entry0, next_entry1;
     grn_bool pushed = GRN_FALSE;
@@ -1663,9 +1665,9 @@ grn_geo_cursor_entry_next(grn_ctx *ctx,
       break;
     }
 
-    memcpy(&next_entry0, entry, sizeof(grn_geo_cursor_entry));
+    grn_memcpy(&next_entry0, entry, sizeof(grn_geo_cursor_entry));
     next_entry0.target_bit++;
-    memcpy(&next_entry1, entry, sizeof(grn_geo_cursor_entry));
+    grn_memcpy(&next_entry1, entry, sizeof(grn_geo_cursor_entry));
     next_entry1.target_bit++;
     SET_N_BIT(next_entry1.key, next_entry1.target_bit);
 
@@ -1767,9 +1769,9 @@ grn_geo_cursor_entry_next(grn_ctx *ctx,
         print_key_mark(ctx, stack_entry->target_bit);
       }
 #endif
-      memcpy(entry,
-             &(area->entries[area->current_entry--]),
-             sizeof(grn_geo_cursor_entry));
+      grn_memcpy(entry,
+                 &(area->entries[area->current_entry--]),
+                 sizeof(grn_geo_cursor_entry));
 #ifdef GEO_DEBUG
       printf("%d: pop entry\n", entry->target_bit);
 #endif
@@ -1786,7 +1788,7 @@ grn_geo_cursor_entry_next(grn_ctx *ctx,
   return GRN_TRUE;
 }
 
-typedef grn_bool (*grn_geo_cursor_callback)(grn_ctx *ctx, grn_ii_posting *posting, void *user_data);
+typedef grn_bool (*grn_geo_cursor_callback)(grn_ctx *ctx, grn_posting *posting, void *user_data);
 
 static void
 grn_geo_cursor_each(grn_ctx *ctx, grn_obj *geo_cursor,
@@ -1797,7 +1799,7 @@ grn_geo_cursor_each(grn_ctx *ctx, grn_obj *geo_cursor,
   grn_table_cursor *pat_cursor;
   grn_ii *ii;
   grn_ii_cursor *ii_cursor;
-  grn_ii_posting *posting = NULL;
+  grn_posting *posting = NULL;
   grn_geo_point *current, *top_left, *bottom_right;
   grn_id index_id;
 
@@ -1884,10 +1886,10 @@ grn_geo_cursor_each(grn_ctx *ctx, grn_obj *geo_cursor,
 }
 
 static grn_bool
-grn_geo_cursor_next_callback(grn_ctx *ctx, grn_ii_posting *posting,
+grn_geo_cursor_next_callback(grn_ctx *ctx, grn_posting *posting,
                              void *user_data)
 {
-  grn_ii_posting **return_posting = user_data;
+  grn_posting **return_posting = user_data;
   *return_posting = posting;
   return GRN_FALSE;
 }
@@ -1895,7 +1897,7 @@ grn_geo_cursor_next_callback(grn_ctx *ctx, grn_ii_posting *posting,
 grn_posting *
 grn_geo_cursor_next(grn_ctx *ctx, grn_obj *geo_cursor)
 {
-  grn_ii_posting *posting = NULL;
+  grn_posting *posting = NULL;
   grn_geo_cursor_each(ctx, geo_cursor, grn_geo_cursor_next_callback, &posting);
   return (grn_posting *)posting;
 }
@@ -1923,7 +1925,7 @@ typedef struct {
 } grn_geo_select_in_rectangle_data;
 
 static grn_bool
-grn_geo_select_in_rectangle_callback(grn_ctx *ctx, grn_ii_posting *posting,
+grn_geo_select_in_rectangle_callback(grn_ctx *ctx, grn_posting *posting,
                                      void *user_data)
 {
   grn_geo_select_in_rectangle_data *data = user_data;
@@ -1979,7 +1981,7 @@ geo_point_get(grn_ctx *ctx, grn_obj *pat, int flags, grn_geo_point *geo_point)
     void *key;
     int key_size;
     key_size = grn_table_cursor_get_key(ctx, cursor, &key);
-    memcpy(geo_point, key, key_size);
+    grn_memcpy(geo_point, key, key_size);
   }
 
 exit:
@@ -1989,13 +1991,13 @@ exit:
   return rc;
 }
 
-int
-grn_geo_estimate_in_rectangle(grn_ctx *ctx,
-                              grn_obj *index,
-                              grn_obj *top_left_point,
-                              grn_obj *bottom_right_point)
+uint32_t
+grn_geo_estimate_size_in_rectangle(grn_ctx *ctx,
+                                   grn_obj *index,
+                                   grn_obj *top_left_point,
+                                   grn_obj *bottom_right_point)
 {
-  int n = 0;
+  uint32_t n = 0;
   int total_records;
   grn_rc rc;
   in_rectangle_data data;
@@ -2004,7 +2006,6 @@ grn_geo_estimate_in_rectangle(grn_ctx *ctx,
   GRN_VOID_INIT(&(data.bottom_right_point_buffer));
   if (in_rectangle_data_prepare(ctx, index, top_left_point, bottom_right_point,
                                 "grn_geo_estimate_in_rectangle()", &data)) {
-    n = -1;
     goto exit;
   }
 
@@ -2025,8 +2026,6 @@ grn_geo_estimate_in_rectangle(grn_ctx *ctx,
       if (rc == GRN_END_OF_DATA) {
         n = total_records;
         rc = GRN_SUCCESS;
-      } else {
-        n = -1;
       }
       goto exit;
     }
@@ -2052,7 +2051,7 @@ grn_geo_estimate_in_rectangle(grn_ctx *ctx,
                        (double)total_longitude_distance);
     }
     estimated_n_records = ceil(total_records * select_ratio);
-    n = (int)estimated_n_records;
+    n = (uint32_t)estimated_n_records;
   }
 
 exit :
@@ -2061,6 +2060,25 @@ exit :
   return n;
 }
 
+int
+grn_geo_estimate_in_rectangle(grn_ctx *ctx,
+                              grn_obj *index,
+                              grn_obj *top_left_point,
+                              grn_obj *bottom_right_point)
+{
+  uint32_t size;
+
+  size = grn_geo_estimate_size_in_rectangle(ctx,
+                                            index,
+                                            top_left_point,
+                                            bottom_right_point);
+  if (ctx->rc != GRN_SUCCESS) {
+    return -1;
+  }
+
+  return size;
+}
+
 grn_bool
 grn_geo_in_circle(grn_ctx *ctx, grn_obj *point, grn_obj *center,
                   grn_obj *radius_or_point,
