@@ -21,6 +21,7 @@
 #include <ctype.h>
 #include <fcntl.h>
 #include <sys/stat.h>
+#include <errno.h>
 
 #ifdef WIN32
 # define GROONGA_MAIN
@@ -32,6 +33,7 @@
 #include <grn_proc.h>
 #include <grn_db.h>
 #include <grn_util.h>
+#include <grn_error.h>
 
 #ifdef HAVE_SYS_WAIT_H
 # include <sys/wait.h>
@@ -103,6 +105,7 @@ static int ready_notify_pipe[2];
 static grn_encoding encoding;
 static grn_command_version default_command_version;
 static int64_t default_match_escalation_threshold;
+static const char *windows_event_source_name = "Groonga";
 static grn_bool use_windows_event_log = GRN_FALSE;
 
 static int
@@ -292,6 +295,11 @@ s_output_raw(grn_ctx *ctx, int flags, FILE *stream)
   if (flags & GRN_CTX_TAIL) {
     grn_obj *command;
 
+    if (grn_ctx_get_output_type(ctx) == GRN_CONTENT_GROONGA_COMMAND_LIST &&
+        chunk_size > 0 &&
+        chunk[chunk_size - 1] != '\n') {
+      fwrite("\n", 1, 1, stream);
+    }
     fflush(stream);
 
     command = GRN_CTX_USER_DATA(ctx)->ptr;
@@ -357,9 +365,6 @@ do_alone(int argc, char **argv)
   grn_obj *db;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
-  if (use_windows_event_log) {
-    grn_windows_event_logger_set(ctx);
-  }
   if (argc > 0 && argv) { path = *argv++; argc--; }
   db = (newdb || !path) ? grn_db_create(ctx, path, NULL) : grn_db_open(ctx, path);
   if (db) {
@@ -437,9 +442,6 @@ g_client(int argc, char **argv)
   const char *hostname = DEFAULT_DEST;
   if (argc > 0 && argv) { hostname = *argv++; argc--; }
   grn_ctx_init(ctx, 0);
-  if (use_windows_event_log) {
-    grn_windows_event_logger_set(ctx);
-  }
   if (!grn_ctx_connect(ctx, hostname, port, 0)) {
     if (!argc) {
       grn_obj text;
@@ -478,25 +480,25 @@ static grn_cond q_cond;
 static uint32_t nthreads = 0, nfthreads = 0, max_nfthreads;
 
 static uint32_t
-groonga_get_thread_count(void *data)
+groonga_get_thread_limit(void *data)
 {
   return max_nfthreads;
 }
 
 static void
-groonga_set_thread_count(uint32_t new_count, void *data)
+groonga_set_thread_limit(uint32_t new_limit, void *data)
 {
   uint32_t i;
   uint32_t current_nfthreads;
 
-  MUTEX_LOCK(q_mutex);
+  MUTEX_LOCK_ENSURE(&grn_gctx, q_mutex);
   current_nfthreads = nfthreads;
-  max_nfthreads = new_count;
+  max_nfthreads = new_limit;
   MUTEX_UNLOCK(q_mutex);
 
-  if (current_nfthreads > new_count) {
+  if (current_nfthreads > new_limit) {
     for (i = 0; i < current_nfthreads; i++) {
-      MUTEX_LOCK(q_mutex);
+      MUTEX_LOCK_ENSURE(&grn_gctx, q_mutex);
       COND_SIGNAL(q_cond);
       MUTEX_UNLOCK(q_mutex);
     }
@@ -544,6 +546,13 @@ create_pid_file(void)
   }
 
   pid_file = fopen(pid_file_path, "w");
+  if (!pid_file) {
+    fprintf(stderr,
+            "Failed to open PID file: <%s>: <%s>\n",
+            pid_file_path, grn_strerror(errno));
+    return;
+  }
+
   {
 #ifdef WIN32
     DWORD pid;
@@ -648,7 +657,7 @@ run_server_loop(grn_ctx *ctx, grn_com_event *ev)
     /* todo : log stat */
   }
   for (;;) {
-    MUTEX_LOCK(q_mutex);
+    MUTEX_LOCK_ENSURE(ctx, q_mutex);
     if (nthreads == nfthreads) { break; }
     MUTEX_UNLOCK(q_mutex);
     grn_nanosleep(1000000);
@@ -1933,12 +1942,9 @@ h_worker(void *arg)
   ht_context hc;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
-  if (use_windows_event_log) {
-    grn_windows_event_logger_set(ctx);
-  }
   grn_ctx_use(ctx, (grn_obj *)arg);
   grn_ctx_recv_handler_set(ctx, h_output, &hc);
-  MUTEX_LOCK(q_mutex);
+  MUTEX_LOCK_ENSURE(ctx, q_mutex);
   GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread start (%d/%d)", nfthreads, nthreads);
   do {
     grn_obj *msg;
@@ -1960,7 +1966,7 @@ h_worker(void *arg)
     hc.in_body = GRN_FALSE;
     hc.is_chunked = GRN_FALSE;
     do_htreq(ctx, (grn_msg *)msg);
-    MUTEX_LOCK(q_mutex);
+    MUTEX_LOCK_ENSURE(ctx, q_mutex);
   } while (nfthreads < max_nfthreads && grn_gctx.stat != GRN_CTX_QUIT);
 exit :
   nthreads--;
@@ -1983,7 +1989,7 @@ h_handler(grn_ctx *ctx, grn_obj *msg)
     /* if not keep alive connection */
     grn_com_event_del(ctx, com->ev, fd);
     ((grn_msg *)msg)->u.fd = fd;
-    MUTEX_LOCK(q_mutex);
+    MUTEX_LOCK_ENSURE(ctx, q_mutex);
     grn_com_queue_enque(ctx, &ctx_new, (grn_com_queue_entry *)msg);
     if (!nfthreads && nthreads < max_nfthreads) {
       grn_thread thread;
@@ -2001,12 +2007,6 @@ h_server(char *path)
   int exit_code = EXIT_FAILURE;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
-  if (use_windows_event_log) {
-    grn_windows_event_logger_set(ctx);
-  }
-  MUTEX_INIT(q_mutex);
-  COND_INIT(q_cond);
-  CRITICAL_SECTION_INIT(cache_lock);
   GRN_COM_QUEUE_INIT(&ctx_new);
   GRN_COM_QUEUE_INIT(&ctx_old);
   check_rlimit_nofile(ctx);
@@ -2018,7 +2018,7 @@ h_server(char *path)
 static grn_thread_func_result CALLBACK
 g_worker(void *arg)
 {
-  MUTEX_LOCK(q_mutex);
+  MUTEX_LOCK_ENSURE(NULL, q_mutex);
   GRN_LOG(&grn_gctx, GRN_LOG_NOTICE, "thread start (%d/%d)", nfthreads, nthreads);
   do {
     grn_ctx *ctx;
@@ -2065,7 +2065,7 @@ g_worker(void *arg)
         while ((msg = (grn_obj *)grn_com_queue_deque(ctx, &edge->send_old))) {
           grn_msg_close(ctx, msg);
         }
-        MUTEX_LOCK(q_mutex);
+        MUTEX_LOCK_ENSURE(ctx, q_mutex);
         if (ctx->stat == GRN_CTX_QUIT || edge->stat == EDGE_ABORT) { break; }
       }
     }
@@ -2086,7 +2086,7 @@ exit :
 static void
 g_dispatcher(grn_ctx *ctx, grn_edge *edge)
 {
-  MUTEX_LOCK(q_mutex);
+  MUTEX_LOCK_ENSURE(ctx, q_mutex);
   if (edge->stat == EDGE_IDLE) {
     grn_com_queue_enque(ctx, &ctx_new, (grn_com_queue_entry *)edge);
     edge->stat = EDGE_WAIT;
@@ -2127,7 +2127,7 @@ g_handler(grn_ctx *ctx, grn_obj *msg)
   if (ctx->rc) {
     if (com->has_sid) {
       if ((edge = com->opaque)) {
-        MUTEX_LOCK(q_mutex);
+        MUTEX_LOCK_ENSURE(ctx, q_mutex);
         if (edge->stat == EDGE_IDLE) {
           grn_com_queue_enque(ctx, &ctx_old, (grn_com_queue_entry *)edge);
         }
@@ -2143,9 +2143,6 @@ g_handler(grn_ctx *ctx, grn_obj *msg)
     edge = grn_edges_add(ctx, &((grn_msg *)msg)->edge_id, &added);
     if (added) {
       grn_ctx_init(&edge->ctx, 0);
-      if (use_windows_event_log) {
-        grn_windows_event_logger_set(&edge->ctx);
-      }
       GRN_COM_QUEUE_INIT(&edge->recv_new);
       GRN_COM_QUEUE_INIT(&edge->send_old);
       grn_ctx_use(&edge->ctx, (grn_obj *)com->ev->opaque);
@@ -2172,12 +2169,6 @@ g_server(char *path)
   int exit_code = EXIT_FAILURE;
   grn_ctx ctx_, *ctx = &ctx_;
   grn_ctx_init(ctx, 0);
-  if (use_windows_event_log) {
-    grn_windows_event_logger_set(ctx);
-  }
-  MUTEX_INIT(q_mutex);
-  COND_INIT(q_cond);
-  CRITICAL_SECTION_INIT(cache_lock);
   GRN_COM_QUEUE_INIT(&ctx_new);
   GRN_COM_QUEUE_INIT(&ctx_old);
   check_rlimit_nofile(ctx);
@@ -2420,7 +2411,7 @@ config_file_load(const char *path, const grn_str_getopt_opt *opts, int *flags)
 static const int default_http_port = DEFAULT_HTTP_PORT;
 static const int default_gqtp_port = DEFAULT_GQTP_PORT;
 static grn_encoding default_encoding = GRN_ENC_DEFAULT;
-static uint32_t default_max_num_threads = DEFAULT_MAX_NFTHREADS;
+static uint32_t default_max_n_threads = DEFAULT_MAX_NFTHREADS;
 static const grn_log_level default_log_level = GRN_LOG_DEFAULT_LEVEL;
 static const char * const default_protocol = "gqtp";
 static const char *default_hostname = "localhost";
@@ -2465,9 +2456,9 @@ init_default_settings(void)
   default_encoding = grn_encoding_parse(GRN_DEFAULT_ENCODING);
 
   {
-    const uint32_t num_cores = get_core_number();
-    if (num_cores != 0) {
-      default_max_num_threads = num_cores;
+    const uint32_t n_cores = get_core_number();
+    if (n_cores != 0) {
+      default_max_n_threads = n_cores;
     }
   }
 
@@ -2699,7 +2690,7 @@ show_usage(FILE *output)
           grn_encoding_to_string(default_encoding),
           default_gqtp_port, default_bind_address,
           default_http_port, default_gqtp_port, default_hostname, default_protocol,
-          default_document_root, default_cache_limit, default_max_num_threads,
+          default_document_root, default_cache_limit, default_max_n_threads,
           grn_log_level_to_string(default_log_level),
           default_log_path, default_query_log_path,
           default_config_path, default_default_command_version,
@@ -2712,7 +2703,7 @@ main(int argc, char **argv)
 {
   const char *port_arg = NULL;
   const char *encoding_arg = NULL;
-  const char *max_num_threads_arg = NULL;
+  const char *max_n_threads_arg = NULL;
   const char *log_level_arg = NULL;
   const char *bind_address_arg = NULL;
   const char *hostname_arg = NULL;
@@ -2769,7 +2760,7 @@ main(int argc, char **argv)
   };
   opts[0].arg = &port_arg;
   opts[1].arg = &encoding_arg;
-  opts[2].arg = &max_num_threads_arg;
+  opts[2].arg = &max_n_threads_arg;
   opts[7].arg = &log_level_arg;
   opts[8].arg = &hostname_arg;
   opts[10].arg = &protocol_arg;
@@ -2944,7 +2935,7 @@ main(int argc, char **argv)
 #endif /* WIN32 */
 
   if (use_windows_event_log) {
-    grn_windows_event_logger_set(NULL);
+    grn_windows_event_logger_set(NULL, windows_event_source_name);
   }
 
   if (log_path_arg) {
@@ -3008,22 +2999,26 @@ main(int argc, char **argv)
     grn_default_logger_set_max_level(log_level);
   }
 
-  if (max_num_threads_arg) {
-    const char * const end = max_num_threads_arg + strlen(max_num_threads_arg);
+  if (max_n_threads_arg) {
+    const char * const end = max_n_threads_arg + strlen(max_n_threads_arg);
     const char *rest = NULL;
-    const uint32_t value = grn_atoui(max_num_threads_arg, end, &rest);
+    const uint32_t value = grn_atoui(max_n_threads_arg, end, &rest);
     if (end != rest || value < 1 || value > 100) {
       fprintf(stderr, "invalid max number of threads: <%s>\n",
-              max_num_threads_arg);
+              max_n_threads_arg);
       return EXIT_FAILURE;
     }
     max_nfthreads = value;
   } else {
-    max_nfthreads = default_max_num_threads;
+    if (flags & FLAG_MODE_ALONE) {
+      max_nfthreads = 1;
+    } else {
+      max_nfthreads = default_max_n_threads;
+    }
   }
 
-  grn_thread_set_get_count_func(groonga_get_thread_count, NULL);
-  grn_thread_set_set_count_func(groonga_set_thread_count, NULL);
+  grn_thread_set_get_limit_func(groonga_get_thread_limit, NULL);
+  grn_thread_set_set_limit_func(groonga_set_thread_limit, NULL);
 
   if (input_path) {
     if (!freopen(input_path, "r", stdin)) {
@@ -3166,7 +3161,11 @@ main(int argc, char **argv)
     line_editor_init(argc, argv);
   }
 #endif
-  if (grn_init()) { return EXIT_FAILURE; }
+  grn_gctx.errbuf[0] = '\0';
+  if (grn_init()) {
+    fprintf(stderr, "failed to initialize Groonga: %s\n", grn_gctx.errbuf);
+    return EXIT_FAILURE;
+  }
 
   grn_set_default_encoding(encoding);
 
@@ -3188,6 +3187,10 @@ main(int argc, char **argv)
     grn_cache_set_max_n_entries(&grn_gctx, cache, cache_limit);
   }
 
+  MUTEX_INIT(q_mutex);
+  COND_INIT(q_cond);
+  CRITICAL_SECTION_INIT(cache_lock);
+
   newdb = (flags & FLAG_NEW_DB);
   is_daemon_mode = (flags & FLAG_MODE_DAEMON);
   if (flags & FLAG_MODE_CLIENT) {
@@ -3198,6 +3201,10 @@ main(int argc, char **argv)
     exit_code = do_alone(argc - i, argv + i);
   }
 
+  CRITICAL_SECTION_FIN(cache_lock);
+  COND_FIN(q_cond);
+  MUTEX_FIN(q_mutex);
+
 #ifdef GRN_WITH_LIBEDIT
   if (need_line_editor) {
     line_editor_fin();
